from argparse import ArgumentParser
from pathlib import Path
import argparse
import glob
import json
import os
import random
import string
import subprocess
import sys
import time

from ramalama.huggingface import Huggingface
from ramalama.common import in_container, container_manager, exec_cmd, run_cmd, default_image, find_working_directory, perror
from ramalama.oci import OCI
from ramalama.ollama import Ollama
from ramalama.shortnames import Shortnames
from ramalama.version import version, print_version


class HelpException(Exception):
    pass


def use_container():
    transport = os.getenv("RAMALAMA_IN_CONTAINER")
    if transport:
        return transport.lower() == "true"

    if in_container() or sys.platform == "darwin":
        return False

    return True


def init_cli():
    parser = ArgumentParser(
        prog="ramalama",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        description="Simple management tool for working with AI Models.",
    )
    parser.add_argument("--store", default=get_store(), help="store AI Models in the specified directory")
    parser.add_argument(
        "--dryrun", dest="dryrun", action="store_true", help="show container runtime command without executing it"
    )
    parser.add_argument("--dry-run", dest="dryrun", action="store_true", help=argparse.SUPPRESS)
    parser.add_argument(
        "--container",
        dest="container",
        default=use_container(),
        action="store_true",
        help="run RamaLama in the default container",
    )
    parser.add_argument(
        "--image",
        default=default_image(),
        help="OCI container image to run with specified AI model",
    )
    parser.add_argument(
        "--nocontainer",
        dest="container",
        default=False,
        action="store_false",
        help="do not run RamaLama in the default container",
    )
    parser.add_argument(
        "--runtime",
        default="llama.cpp",
        choices=["llama.cpp", "vllm"],
        help="specify the runtime to use, valid options are 'llama.cpp' and 'vllm'",
    )
    parser.add_argument("-v", dest="version", action="store_true", help="show RamaLama version")

    subparsers = parser.add_subparsers(dest="subcommand")
    subparsers.required = False

    help_parser(subparsers)
    containers_parser(subparsers)
    list_parser(subparsers)
    login_parser(subparsers)
    logout_parser(subparsers)
    pull_parser(subparsers)
    push_parser(subparsers)
    rm_parser(subparsers)
    run_parser(subparsers)
    serve_parser(subparsers)
    stop_parser(subparsers)
    version_parser(subparsers)
    # Parse CLI
    args = parser.parse_args()

    # create stores directories
    mkdirs(args.store)
    if hasattr(args, "MODEL"):
        shortnames = Shortnames()
        resolved_model = shortnames.resolve(args.MODEL)
        if resolved_model:
            args.UNRESOLVED_MODEL = args.MODEL
            args.MODEL = resolved_model

    return parser, args


def login_parser(subparsers):
    parser = subparsers.add_parser("login", help="login to remote registry")
    # Do not run in a container
    parser.add_argument("--container", default=False, action="store_false", help=argparse.SUPPRESS)
    parser.add_argument("-p", "--password", dest="password", help="password for registry")
    parser.add_argument(
        "--password-stdin", dest="passwordstdin", action="store_true", help="take the password for registry from stdin"
    )
    parser.add_argument("--token", dest="token", help="token for registry")
    parser.add_argument("-u", "--username", dest="username", help="username for registry")
    parser.add_argument("TRANSPORT", nargs="?", type=str, default="")  # positional argument
    parser.set_defaults(func=login_cli)


def login_cli(args):
    transport = args.TRANSPORT
    if transport != "":
        transport = os.getenv("RAMALAMA_TRANSPORT")
    model = New(str(transport))
    return model.login(args)


def logout_parser(subparsers):
    parser = subparsers.add_parser("logout", help="logout from remote registry")
    # Do not run in a container
    parser.add_argument("--container", default=False, action="store_false", help=argparse.SUPPRESS)
    parser.add_argument("--token", help="token for registry")
    parser.add_argument("TRANSPORT", nargs="?", type=str, default="")  # positional argument
    parser.add_argument("TRANSPORT", nargs="?", type=str, default="")  # positional argument
    parser.set_defaults(func=logout_cli)


def logout_cli(args):
    transport = args.TRANSPORT
    if transport != "":
        transport = os.getenv("RAMALAMA_TRANSPORT")
    model = New(str(transport))
    return model.logout(args)


def mkdirs(store):
    # List of directories to create
    directories = [
        "models/huggingface",
        "repos/huggingface",
        "models/oci",
        "repos/oci",
        "models/ollama",
        "repos/ollama",
    ]

    # Create each directory
    for directory in directories:
        full_path = os.path.join(store, directory)
        os.makedirs(full_path, exist_ok=True)


def human_duration(d):
    if d < 1:
        return "Less than a second"
    elif d == 1:
        return "1 second"
    elif d < 60:
        return f"{d} seconds"
    elif d < 120:
        return "1 minute"
    elif d < 3600:
        return f"{d // 60} minutes"
    elif d < 7200:
        return "1 hour"
    elif d < 86400:
        return f"{d // 3600} hours"
    elif d < 172800:
        return "1 day"
    elif d < 604800:
        return f"{d // 86400} days"
    elif d < 1209600:
        return "1 week"
    elif d < 2419200:
        return f"{d // 604800} weeks"
    elif d < 4838400:
        return "1 month"
    elif d < 31536000:
        return f"{d // 2419200} months"
    elif d < 63072000:
        return "1 year"
    else:
        return f"{d // 31536000} years"


def list_files_by_modification():
    return sorted(Path().rglob("*"), key=lambda p: os.path.getmtime(p), reverse=True)


def containers_parser(subparsers):
    parser = subparsers.add_parser("containers", aliases=["ps"], help="list all RamaLama containers")
    parser.add_argument("--format", help="pretty-print containers to JSON or using a Go template")
    parser.add_argument("-n", "--noheading", dest="noheading", action="store_true", help="do not display heading")
    parser.add_argument("--no-trunc", dest="notrunc", action="store_true", help="display the extended information")
    parser.add_argument("--container", default=False, action="store_false", help=argparse.SUPPRESS)
    parser.set_defaults(func=list_containers)


def _list_containers(args):
    conman = container_manager()
    if conman == "":
        raise IndexError("no container manager (Podman, Docker) found")

    conman_args = [conman, "ps", "-a", "--filter", "label=RAMALAMA container"]
    if args.noheading:
        conman_args += ["--noheading"]
    if hasattr(args, "notrunc") and args.notrunc:
        conman_args += ["--no-trunc"]

    if args.format:
        conman_args += [f"--format={args.format}"]

    try:
        output = run_cmd(conman_args).stdout.decode("utf-8").strip()
        if output == "":
            return []
        return output.split("\n")
    except subprocess.CalledProcessError as e:
        perror("ramalama list command requires a running container engine")
        raise(e)

def list_containers(args):
    if len(_list_containers(args)) == 0:
        return
    print("\n".join(_list_containers(args)))


def list_parser(subparsers):
    parser = subparsers.add_parser("list", aliases=["ls"], help="list all downloaded AI Models")
    parser.add_argument("-n", "--noheading", dest="noheading", action="store_true", help="do not display heading")
    parser.add_argument("--json", dest="json", action="store_true", help="print using json")
    parser.add_argument("-q", "--quiet", dest="quiet", action="store_true", help="print only Model names")
    parser.set_defaults(func=list_cli)


def human_readable_size(size):
    for unit in ["B", "KB", "MB", "GB", "TB"]:
        if size < 1024:
            size = round(size, 2)
            return f"{size} {unit}"

        size /= 1024

    return f"{size} PB"


def get_size(file):
    return human_readable_size(os.path.getsize(file))


def _list_models(args):
    mycwd = os.getcwd()
    os.chdir(f"{args.store}/models/")
    models = []

    # Collect model data
    for path in list_files_by_modification():
        if path.is_symlink():
            name = str(path).replace("/", "://", 1)
            file_epoch = path.lstat().st_mtime
            modified = int(time.time() - file_epoch)
            size = get_size(path)

            # Store data for later use
            models.append({"name": name, "modified": modified, "size": size})

    os.chdir(mycwd)
    return models


def list_cli(args):
    models = _list_models(args)

    # If JSON output is requested
    if args.json:
        print(json.dumps(models))
        return

    # Calculate maximum width for each column
    name_width = len("NAME")
    modified_width = len("MODIFIED")
    size_width = len("SIZE")
    for model in models:
        modified = human_duration(model["modified"]) + " ago"
        name_width = max(name_width, len(model["name"]))
        modified_width = max(modified_width, len(modified))
        size_width = max(size_width, len(model["size"]))

    if not args.quiet and not args.noheading and not args.json:
        print(f"{'NAME':<{name_width}} {'MODIFIED':<{modified_width}} {'SIZE':<{size_width}}")

    for model in models:
        if args.quiet:
            print(model["name"])
        else:
            print(f"{model['name']:<{name_width}} {modified:<{modified_width}} {model['size']:<{size_width}}")


def help_parser(subparsers):
    parser = subparsers.add_parser("help", help="help about any command")
    # Do not run in a container
    parser.add_argument("--container", default=False, action="store_false", help=argparse.SUPPRESS)
    parser.set_defaults(func=help_cli)


def help_cli(args):
    raise HelpException()


def pull_parser(subparsers):
    parser = subparsers.add_parser("pull", help="pull AI Model from Model registry to local storage")
    parser.add_argument("MODEL")  # positional argument
    parser.set_defaults(func=pull_cli)


def pull_cli(args):
    model = New(args.MODEL)
    matching_files = glob.glob(f"{args.store}/models/*/{model}")
    if matching_files:
        return matching_files[0]

    return model.pull(args)


def push_parser(subparsers):
    parser = subparsers.add_parser("push", help="push AI Model from local storage to remote registry")
    parser.add_argument("SOURCE")  # positional argument
    parser.add_argument("TARGET")  # positional argument
    parser.set_defaults(func=push_cli)


def push_cli(args):
    smodel = New(args.SOURCE)
    source = smodel.path(args)

    model = New(args.TARGET)
    model.push(source, args)


def _name():
    return "ramalama_" + "".join(random.choices(string.ascii_letters + string.digits, k=10))


def run_parser(subparsers):
    parser = subparsers.add_parser("run", help="run specified AI Model as a chatbot")
    parser.add_argument("-n", "--name", dest="name", help="name of container in which the Model will be run")
    parser.add_argument("MODEL")  # positional argument
    parser.add_argument("ARGS", nargs="*", help="Overrides the default prompt, and the output is returned without entering the chatbot")
    parser.set_defaults(func=run_cli)


def run_cli(args):
    model = New(args.MODEL)
    model.run(args)


def serve_parser(subparsers):
    parser = subparsers.add_parser("serve", help="serve REST API on specified AI Model")
    parser.add_argument(
        "-d", "--detach", action="store_true", default=True, dest="detach", help="run the container in detached mode"
    )
    parser.add_argument("-n", "--name", dest="name", help="name of container in which the Model will be run")
    parser.add_argument("-p", "--port", default="8080", help="port for AI Model server to listen on")
    parser.add_argument(
        "--generate",
        choices=["quadlet"],
        help="generate specified configuration format for running the AI Model as a service",
    )
    parser.add_argument("MODEL")  # positional argument
    parser.set_defaults(func=serve_cli)


def serve_cli(args):
    if not args.container:
        args.detach = False
    model = New(args.MODEL)
    model.serve(args)


def stop_parser(subparsers):
    parser = subparsers.add_parser("stop", help="stop named container that is running AI Model")
    parser.add_argument("--container", default=False, action="store_false", help=argparse.SUPPRESS)
    parser.add_argument("-a", "--all", action="store_true", help="stop all RamaLama containers")
    parser.add_argument(
        "--ignore", action="store_true", help="ignore errors when specified RamaLama container is missing"
    )
    parser.add_argument("NAME", nargs="?")  # positional argument
    parser.set_defaults(func=stop_container)


def _stop_container(args, name):
    if not name:
        raise IndexError("must specify a container name")
    conman = container_manager()
    if conman == "":
        raise IndexError("no container manager (Podman, Docker) found")

    conman_args = [conman, "stop", "-t=0"]
    if args.ignore:
        conman_args += [ "--ignore", str(args.ignore)]
    conman_args += [name]
    run_cmd(conman_args)


def stop_container(args):
    if not args.all:
        return _stop_container(args, args.NAME)

    if args.NAME:
        raise IndexError("specifying --all and container name, %s, not allowed" % args.NAME)
    args.noheading = True
    args.ignore = True
    args.format = "{{ .Names }}"
    for i in _list_containers(args):
        _stop_container(args, i)


def version_parser(subparsers):
    parser = subparsers.add_parser("version", help="display version of AI Model")
    # Do not run in a container
    parser.add_argument("--container", default=False, action="store_false", help=argparse.SUPPRESS)
    parser.set_defaults(func=print_version)


def rm_parser(subparsers):
    parser = subparsers.add_parser("rm", help="remove AI Model from local storage")
    parser.add_argument("-a", "--all", action="store_true", help="remove all local Models")
    parser.add_argument("--ignore", action="store_true", help="ignore errors when specified Model does not exist")
    parser.add_argument("MODEL", nargs="?")
    parser.set_defaults(func=rm_cli)


def _rm_model(model, args):
    model = New(model)
    model.remove(args)


def rm_cli(args):
    if not args.all:
        return _rm_model(args.MODEL, args)

    if args.MODEL == "":
        raise IndexError("can not specify --all as well MODEL")

    args.noheading = True
    for model in _list_models(args):
        _rm_model(model["name"], args)


def get_store():
    if os.geteuid() == 0:
        return "/var/lib/ramalama"

    return os.path.expanduser("~/.local/share/ramalama")


def run_container(args):
    if hasattr(args, "generate") and args.generate:
        return False

    if not args.container:
        if hasattr(args, "name") and args.name:
            raise IndexError("--nocontainer and --name options conflict. --name requires a container.")

        # --nocontainer implies --detach=false
        if hasattr(args, "detach"):
            args.detach = False
        return False

    if in_container():
        return False

    conman = container_manager()
    if conman == "":
        return False

    if hasattr(args, "name") and args.name:
        name = args.name
    else:
        name = _name()

    wd = find_working_directory()
    conman_args = [
        conman,
        "run",
        "--rm",
        "-i",
        "--label",
        "RAMALAMA container",
        "--security-opt=label=disable",
        "-e",
        "RAMALAMA_TRANSPORT",
        "--name",
        name,
        f"-v{args.store}:/var/lib/ramalama",
        f"-v{os.path.realpath(sys.argv[0])}:/usr/bin/ramalama:ro",
        f"-v{wd}:/usr/share/ramalama/ramalama:ro",
    ]

    di_volume = distinfo_volume()
    if di_volume != "":
        conman_args += [di_volume]

    if sys.stdout.isatty():
        conman_args += ["-t"]

    if hasattr(args, "detach") and args.detach is True:
        conman_args += ["-d"]

    if hasattr(args, "port"):
        conman_args += ["-p", f"{args.port}:{args.port}"]

    if os.path.exists("/dev/dri"):
        conman_args += ["--device", "/dev/dri"]

    if os.path.exists("/dev/kfd"):
        conman_args += ["--device", "/dev/kfd"]

    conman_args += [args.image, "python3", "/usr/bin/ramalama"]
    conman_args += sys.argv[1:]
    if hasattr(args, "UNRESOLVED_MODEL"):
        index = conman_args.index(args.UNRESOLVED_MODEL)
        conman_args[index] = args.MODEL

    if args.dryrun:
        dry_run(conman_args)
        return True

    exec_cmd(conman_args)


def dry_run(args):
    for arg in args:
        if not arg:
            continue
        if " " in arg:
            print('"%s"' % arg, end=" ")
        else:
            print("%s" % arg, end=" ")
    print()


def New(model):
    if model.startswith("huggingface"):
        return Huggingface(model)
    if model.startswith("ollama"):
        return Ollama(model)
    if model.startswith("oci") | model.startswith("docker"):
        return OCI(model)

    transport = os.getenv("RAMALAMA_TRANSPORT")
    if transport == "huggingface":
        return Huggingface(model)
    if transport == "ollama":
        return Ollama(model)
    if transport == "oci":
        return OCI(model)

    return Ollama(model)


def distinfo_volume():
    dist_info = "ramalama-%s.dist-info" % version()
    path = os.path.join(os.path.dirname(os.path.dirname(__file__)), dist_info)
    if not os.path.exists(path):
        return ""

    return f"-v{path}:/usr/share/ramalama/{dist_info}:ro"



import os
import urllib.request
import json
from ramalama.common import run_cmd
from ramalama.model import Model

bar_format = "Pulling {desc}: {percentage:3.0f}% ▕{bar:20}▏ {n_fmt}/{total_fmt} {rate_fmt} {remaining}"


def download_file(url, dest_path, headers=None):
    try:
        from tqdm import tqdm
    except FileNotFoundError:
        raise NotImplementedError(
            """\
Ollama models requires the tqdm modules.
This model can be installed via PyPi tools like pip, pip3, pipx or via
distribution package managers like dnf or apt. Example:
pip install tqdm
"""
        )

    request = urllib.request.Request(url, headers=headers or {})
    with urllib.request.urlopen(request) as response:
        total_size = int(response.headers.get("Content-Length", 0))
        chunk_size = 8192  # 8 KB chunks
        with open(dest_path, "wb") as file, tqdm(
            desc=dest_path[-16:],
            total=total_size,
            unit="B",
            unit_scale=True,
            unit_divisor=1024,
            bar_format=bar_format,
            ascii=True,
        ) as progress_bar:
            while True:
                chunk = response.read(chunk_size)
                if not chunk:
                    break
                file.write(chunk)
                progress_bar.update(len(chunk))


def pull_manifest(repos, manifests, accept, registry_head, model_tag):
    os.makedirs(os.path.dirname(manifests), exist_ok=True)
    os.makedirs(os.path.join(repos, "blobs"), exist_ok=True)
    url = f"{registry_head}/manifests/{model_tag}"
    headers = {"Accept": accept}
    download_file(url, manifests, headers=headers)


def pull_config_blob(repos, accept, registry_head, manifest_data):
    cfg_hash = manifest_data["config"]["digest"]
    config_blob_path = os.path.join(repos, "blobs", cfg_hash)
    url = f"{registry_head}/blobs/{cfg_hash}"
    headers = {"Accept": accept}
    download_file(url, config_blob_path, headers=headers)


def pull_blob(repos, layer_digest, accept, registry_head, models, model_name, model_tag, symlink_path):
    layer_blob_path = os.path.join(repos, "blobs", layer_digest)
    url = f"{registry_head}/blobs/{layer_digest}"
    headers = {"Accept": accept}
    download_file(url, layer_blob_path, headers=headers)

    os.makedirs(models, exist_ok=True)
    relative_target_path = os.path.relpath(layer_blob_path, start=os.path.dirname(symlink_path))
    run_cmd(["ln", "-sf", relative_target_path, symlink_path])


def init_pull(repos, manifests, accept, registry_head, model_name, model_tag, models, symlink_path, model):
    try:
        pull_manifest(repos, manifests, accept, registry_head, model_tag)
        with open(manifests, "r") as f:
            manifest_data = json.load(f)
    except urllib.error.HTTPError as e:
        if e.code == 404:
            raise KeyError(f"{model} not found")
        raise e

    pull_config_blob(repos, accept, registry_head, manifest_data)
    for layer in manifest_data["layers"]:
        layer_digest = layer["digest"]
        if layer["mediaType"] != "application/vnd.ollama.image.model":
            continue

        pull_blob(repos, layer_digest, accept, registry_head, models, model_name, model_tag, symlink_path)

    return symlink_path


class Ollama(Model):
    def __init__(self, model):
        super().__init__(model.removeprefix("ollama://"))
        self.type = "Ollama"

    def _local(self, args):
        models = args.store + "/models/ollama"
        if "/" in self.model:
            model_full = self.model
            self._models = os.path.join(models, model_full.rsplit("/", 1)[0])
        else:
            model_full = "library/" + self.model

        if ":" in model_full:
            model_name, model_tag = model_full.split(":", 1)
        else:
            model_name = model_full
            model_tag = "latest"

        model_base = os.path.basename(model_name)
        symlink_path = os.path.join(models, f"{model_base}:{model_tag}")
        return symlink_path, models, model_base, model_name, model_tag

    def path(self, args):
        symlink_path, _, _, _, _ = self._local(args)
        if not os.path.exists(symlink_path):
            raise KeyError("f{{args.Model} does not exist")

        return symlink_path

    def pull(self, args):
        repos = args.store + "/repos/ollama"
        symlink_path, models, model_base, model_name, model_tag = self._local(args)
        if os.path.exists(symlink_path):
            return symlink_path

        registry = "https://registry.ollama.ai"
        accept = "Accept: application/vnd.docker.distribution.manifest.v2+json"
        manifests = os.path.join(repos, "manifests", registry, model_name, model_tag)
        registry_head = f"{registry}/v2/{model_name}"
        return init_pull(
            repos, manifests, accept, registry_head, model_name, model_tag, models, symlink_path, self.model
        )

    def symlink_path(self, args):
        models = args.store + "/models/ollama"
        if "/" in self.model:
            model_full = self.model
            models = os.path.join(models, model_full.rsplit("/", 1)[0])
        else:
            model_full = "library/" + self.model

        if ":" in model_full:
            model_name, model_tag = model_full.split(":", 1)
        else:
            model_name = model_full
            model_tag = "latest"

        model_base = os.path.basename(model_name)
        return os.path.join(models, f"{model_base}:{model_tag}")



import os
import sys
from ramalama.common import container_manager, exec_cmd, default_image


class Model:
    """Model super class"""

    model = ""
    conman = container_manager()
    type = "Model"
    common_params = ["-c", "2048"]

    def __init__(self, model):
        self.model = model
        if sys.platform == "darwin":
            self.common_params += ["-ngl", "99"]

    def login(self, args):
        raise NotImplementedError(f"ramalama login for {self.type} not implemented")

    def logout(self, args):
        raise NotImplementedError(f"ramalama logout for {self.type} not implemented")

    def path(self, source, args):
        raise NotImplementedError(f"ramalama puath for {self.type} not implemented")

    def pull(self, args):
        raise NotImplementedError(f"ramalama pull for {self.type} not implemented")

    def push(self, source, args):
        raise NotImplementedError(f"ramalama push for {self.type} not implemented")

    def is_symlink_to(self, file_path, target_path):
        if os.path.islink(file_path):
            symlink_target = os.readlink(file_path)
            abs_symlink_target = os.path.abspath(os.path.join(os.path.dirname(file_path), symlink_target))
            abs_target_path = os.path.abspath(target_path)
            return abs_symlink_target == abs_target_path

        return False

    def garbage_collection(self, args):
        repo_paths = ["huggingface", "oci", "ollama"]
        for repo in repo_paths:
            repo_dir = f"{args.store}/repos/{repo}"
            model_dir = f"{args.store}/models/{repo}"
            for root, dirs, files in os.walk(repo_dir):
                file_has_a_symlink = False
                for file in files:
                    file_path = os.path.join(root, file)
                    if (repo == "ollama" and file.startswith("sha256:")) or file.endswith(".gguf"):
                        file_path = os.path.join(root, file)
                        for model_root, model_dirs, model_files in os.walk(model_dir):
                            for model_file in model_files:
                                if self.is_symlink_to(os.path.join(root, model_root, model_file), file_path):
                                    file_has_a_symlink = True

                        if not file_has_a_symlink:
                            os.remove(file_path)
                            file_path = os.path.basename(file_path)
                            print(f"Deleted: {file_path}")

    def remove(self, args):
        symlink_path = self.symlink_path(args)
        if os.path.exists(symlink_path):
            try:
                os.remove(symlink_path)
                print(f"Untagged: {self.model}")
            except OSError as e:
                if not args.ignore:
                    raise KeyError(f"removing {self.model}: {e}")
        else:
            if not args.ignore:
                raise KeyError(f"model {self.model} not found")

        self.garbage_collection(args)

    def symlink_path(self, args):
        raise NotImplementedError(f"symlink_path for {self.type} not implemented")

    def run(self, args):
        prompt = "You are a helpful assistant"
        if args.ARGS:
            prompt = " ".join(args.ARGS)

        symlink_path = self.pull(args)
        exec_args = [
            "llama-cli",
            "-m",
            symlink_path,
            "--in-prefix",
            "",
            "--in-suffix",
            "",
            "--no-display-prompt",
            "-p",
            prompt,
        ] + self.common_params
        if not args.ARGS:
            exec_args.append("-cnv")

        exec_cmd(exec_args, False)

    def serve(self, args):
        symlink_path = self.pull(args)
        exec_args = ["llama-server", "--port", args.port, "-m", symlink_path]
        if args.runtime == "vllm":
            exec_args = ["vllm", "serve", "--port", args.port, symlink_path]

        if args.generate == "quadlet":
            return self.quadlet(symlink_path, args, exec_args)

        exec_cmd(exec_args)

    def quadlet(self, model, args, exec_args):
        port_string = ""
        if hasattr(args, "port"):
            port_string = f"PublishPort={args.port}"

        name_string = ""
        if hasattr(args, "name") and args.name:
            name_string = f"ContainerName={args.name}"

        print(
            f"""
[Unit]
Description=RamaLama {args.UNRESOLVED_MODEL} AI Model Service
After=local-fs.target

[Container]
AddDevice=-/dev/dri
AddDevice=-/dev/kfd
Exec={" ".join(exec_args)}
Image={default_image()}
Volume={model}:{model}:ro,z
{name_string}
{port_string}

[Install]
# Start by default on boot
WantedBy=multi-user.target default.target
"""
        )



import os
import subprocess
import sys

from ramalama.model import Model
from ramalama.common import run_cmd, exec_cmd, perror, available


class OCI(Model):
    def __init__(self, model):
        super().__init__(model.removeprefix("oci://").removeprefix("docker://"))
        self.type = "OCI"
        if available("omlmd"):
            self.omlmd = "omlmd"
        else:
            for i in sys.path:
                self.omlmd = f"{i}/../../../bin/omlmd"
                if os.path.exists(self.omlmd):
                    break
            raise """\
OCI models requires the omlmd module.
This module can be installed via PyPi tools like pip, pip3, pipx or via
distribution package managers like dnf or apt. Example:
pip install omlmd
"""

    def login(self, args):
        conman_args = [self.conman, "login"]
        if args.username:
            conman_args.extend(["--username", args.username])
        if args.password:
            conman_args.extend(["--password", args.password])
        if args.passwordstdin:
            conman_args.append("--password-stdin")
        conman_args.append(args.TRANSPORT)
        return exec_cmd(conman_args)

    def logout(self, args):
        conman_args = [self.conman, "logout"]
        conman_args.append(self.model)
        return exec_cmd(conman_args)

    def _target_decompose(self, model):
        # Remove the prefix and extract target details
        try:
            registry, reference = model.split("/", 1)
        except Exception:
            raise KeyError(
                f"You must specify a registry for the model in the form "
                f"'oci://registry.acme.org/ns/repo:tag', got instead: {self.model}"
            )

        reference_dir = reference.replace(":", "/")
        return registry, reference, reference_dir

    def push(self, source, args):
        target = args.TARGET.strip("oci://")
        tregistry, _, treference_dir = self._target_decompose(target)

        try:
            # Push the model using omlmd, using cwd the model's file parent directory

            run_cmd([self.omlmd, "push", target, source, "--empty-metadata"])
        except subprocess.CalledProcessError as e:
            perror(f"Failed to push model to OCI: {e}")
            raise e

    def pull(self, args):
        try:
            registry, reference = self.model.split("/", 1)
        except Exception:
            registry = "docker.io"
            reference = self.model

        reference_dir = reference.replace(":", "/")
        outdir = f"{args.store}/repos/oci/{registry}/{reference_dir}"
        print(f"Downloading {self.model}...")
        # note: in the current way RamaLama is designed, cannot do Helper(OMLMDRegistry()).pull(target, outdir)
        # since cannot use modules/sdk, can use only cli bindings from pip installs
        run_cmd([self.omlmd, "pull", self.model, "--output", outdir])
        ggufs = [file for file in os.listdir(outdir) if file.endswith(".gguf")]
        if len(ggufs) != 1:
            raise KeyError(f"unable to identify .gguf file in: {outdir}")

        directory = f"{args.store}/models/oci/{registry}/{reference_dir}"
        os.makedirs(directory, exist_ok=True)
        symlink_path = f"{directory}/{ggufs[0]}"
        relative_target_path = os.path.relpath(f"{outdir}/{ggufs[0]}", start=os.path.dirname(symlink_path))
        if os.path.exists(symlink_path) and os.readlink(symlink_path) == relative_target_path:
            # Symlink is already correct, no need to update it
            return symlink_path

        run_cmd(["ln", "-sf", relative_target_path, symlink_path])

        return symlink_path

    def symlink_path(self, args):
        registry, reference = self.model.split("/", 1)
        reference_dir = reference.replace(":", "/")
        path = f"{args.store}/models/oci/{registry}/{reference_dir}"

        if os.path.isfile(path):
            return path

        ggufs = [file for file in os.listdir(path) if file.endswith(".gguf")]
        if len(ggufs) != 1:
            raise KeyError(f"unable to identify .gguf file in: {path}")

        return f"{path}/{ggufs[0]}"



import os
from ramalama.common import run_cmd, exec_cmd
from ramalama.model import Model

missing_huggingface = """
Huggingface models requires the huggingface-cli and tqdm modules.
These modules can be installed via PyPi tools like pip, pip3, pipx or via
distribution package managers like dnf or apt. Example:
pip install huggingface_hub[cli] tqdm
"""


class Huggingface(Model):
    def __init__(self, model):
        super().__init__(model.removeprefix("huggingface://"))
        self.type = "HuggingFace"
        split = self.model.rsplit("/", 1)
        self.directory = ""
        if len(split) > 1:
            self.directory = split[0]
            self.filename = split[1]
        else:
            self.filename = split[0]

    def login(self, args):
        conman_args = ["huggingface-cli", "login"]
        if args.token:
            conman_args.extend(["--token", args.token])
        try:
            self.exec(conman_args)
        except FileNotFoundError as e:
            raise NotImplementedError(
                """\
%s
%s"""
                % (str(e).strip("'"), missing_huggingface)
            )

    def logout(self, args):
        conman_args = ["huggingface-cli", "logout"]
        if args.token:
            conman_args.extend(["--token", args.token])
        conman_args.extend(args)
        self.exec(conman_args)

    def path(self, args):
        return self.symlink_path(args)

    def pull(self, args):
        relative_target_path = ""
        symlink_path = self.symlink_path(args)

        gguf_path = self.download(args.store)
        relative_target_path = os.path.relpath(gguf_path.rstrip(), start=os.path.dirname(symlink_path))
        directory = f"{args.store}/models/huggingface/{self.directory}"
        os.makedirs(directory, exist_ok=True)

        if os.path.exists(symlink_path) and os.readlink(symlink_path) == relative_target_path:
            # Symlink is already correct, no need to update it
            return symlink_path

        run_cmd(["ln", "-sf", relative_target_path, symlink_path])

        return symlink_path

    def push(self, source, args):
        try:
            proc = run_cmd(
                [
                    "huggingface-cli",
                    "upload",
                    "--repo-type",
                    "model",
                    self.directory,
                    self.filename,
                    "--cache-dir",
                    args.store + "/repos/huggingface/.cache",
                    "--local-dir",
                    args.store + "/repos/huggingface/" + self.directory,
                ]
            )
            return proc.stdout.decode("utf-8")
        except FileNotFoundError as e:
            raise NotImplementedError(
                """\
                %s
                %s"""
                % (str(e).strip("'"), missing_huggingface)
            )

    def symlink_path(self, args):
        return f"{args.store}/models/huggingface/{self.directory}/{self.filename}"

    def exec(self, args):
        try:
            exec_cmd(args)
        except FileNotFoundError as e:
            raise NotImplementedError(
                """\
%s

%s
"""
                % str(e).strip("'"),
                missing_huggingface,
            )

    def download(self, store):
        try:
            proc = run_cmd(
                [
                    "huggingface-cli",
                    "download",
                    self.directory,
                    self.filename,
                    "--cache-dir",
                    store + "/repos/huggingface/.cache",
                    "--local-dir",
                    store + "/repos/huggingface/" + self.directory,
                ]
            )
            return proc.stdout.decode("utf-8")
        except FileNotFoundError as e:
            raise NotImplementedError(
                """\
                %s
                %s"""
                % (str(e).strip("'"), missing_huggingface)
            )



import setuptools
import os
from setuptools import find_packages
from setuptools.command.build_py import build_py as build_py_orig


def generate_man_pages(share_path, docs):
    data_files = []

    for path, _, files in os.walk(docs):
        list_entry = (share_path, [os.path.join(path, f) for f in files if f.endswith(".1")])
        data_files.append(list_entry)

    return data_files


def generate_completions(share_path, completions):
    data_files = []

    def remove_prefix(s, prefix):
        if s.startswith(prefix):
            length = len(prefix) + 1
            return s[length:]
        else:
            return s

    for path, _, files in os.walk(completions):
        if len(files) == 0:
            continue
        list_entry = (
            os.path.join(share_path, remove_prefix(path, completions)),
            [os.path.join(path, f) for f in files],
        )
        data_files.append(list_entry)
    return data_files


class build_py(build_py_orig):
    def find_package_modules(self, package, package_dir):
        modules = super().find_package_modules(package, package_dir)
        return [(pkg, mod, file) for (pkg, mod, file) in modules]


setuptools.setup(
    packages=find_packages(),
    cmdclass={"build_py": build_py},
    scripts=["bin/ramalama"],
    data_files=[("share/ramalama", ["shortnames/shortnames.conf"])]
    + generate_completions("share", "build/completions")
    + generate_man_pages("share/man/man1", "docs"),
)



"""ramalama common module."""

import hashlib
import os
import shutil
import subprocess
import sys

x = False


def in_container():
    if os.path.exists("/run/.containerenv") or os.path.exists("/.dockerenv") or os.getenv("container"):
        return True

    return False


def container_manager():
    engine = os.getenv("RAMALAMA_CONTAINER_ENGINE")
    if engine:
        return engine

    if available("podman"):
        return "podman"

    if available("docker"):
        return "docker"

    return ""


def perror(*args, **kwargs):
    print(*args, file=sys.stderr, **kwargs)


def available(cmd):
    return shutil.which(cmd) is not None


def exec_cmd(args, stderr=True):
    if x:
        print(*args)

    if not stderr:
        # Redirecting stderr to /dev/null
        with open(os.devnull, "w") as devnull:
            os.dup2(devnull.fileno(), sys.stderr.fileno())

    try:
        return os.execvp(args[0], args)
    except Exception:
        perror(f"os.execvp({args[0]}, {args})")
        raise


def run_cmd(args, cwd=None):
    """
    Run the given command arguments.

    Args:
    args: command line arguments to execute in a subprocess
    cwd: optional working directory to run the command from
    """
    if x:
        print(*args)

    return subprocess.run(args, check=True, cwd=cwd, stdout=subprocess.PIPE)


def find_working_directory():
    return os.path.dirname(__file__)


def run_curl_cmd(args, filename):
    if not verify_checksum(filename):
        try:
            run_cmd(args)
        except subprocess.CalledProcessError as e:
            if e.returncode == 22:
                perror(filename + " not found")
            raise e


def verify_checksum(filename):
    """
    Verifies if the SHA-256 checksum of a file matches the checksum provided in
    the filename.

    Args:
    filename (str): The filename containing the checksum prefix
                    (e.g., "sha256:<checksum>")

    Returns:
    bool: True if the checksum matches, False otherwise.
    """

    if not os.path.exists(filename):
        return False

    # Check if the filename starts with "sha256:"
    fn_base = os.path.basename(filename)
    if not fn_base.startswith("sha256:"):
        raise ValueError(f"filename does not start with 'sha256:': {fn_base}")

    # Extract the expected checksum from the filename
    expected_checksum = fn_base.split(":")[1]
    if len(expected_checksum) != 64:
        raise ValueError("invalid checksum length in filename")

    # Calculate the SHA-256 checksum of the file contents
    sha256_hash = hashlib.sha256()
    with open(filename, "rb") as f:
        for byte_block in iter(lambda: f.read(4096), b""):
            sha256_hash.update(byte_block)

    # Compare the checksums
    return sha256_hash.hexdigest() == expected_checksum


# default_image function should figure out which GPU the system uses t
# then running appropriate container image.
def default_image():
    image = os.getenv("RAMALAMA_IMAGE")
    if image:
        return image
    return "quay.io/ramalama/ramalama:latest"



"""ramalama client module."""

from ramalama.common import perror
from ramalama.cli import init_cli, run_container, print_version, HelpException
import sys

assert sys.version_info >= (3, 6), "Python 3.6 or greater is required."

__all__ = ["perror", "init_cli", "run_container", "print_version", "HelpException"]
